/**
 * @file     AWList.h
 * @date     2020-10-31
 * @author   whiothes <whiothes81 @gmail.com>
 * @version  1.0
 * @brief    list operations
 */

#ifndef AWLIST_H
#define AWLIST_H

/**
 * @brief      node of list
 *
 * @param      next       pointer to next
 * @param      data       data of node
 * @param      delete_func  function to free data
 */
typedef struct _node {
	struct _node *next;
	struct _node *prev;
	void *data;
	void (*delete_func)(void *);
} AWNode;

/**
 * @brief      list structure
 * @param      head     head of list
 * @param      tail     tail of list
 * @param      curr     current position of list
 * @param      free_me  need to free itself or not
 */
typedef struct _linked_list {
	AWNode *head; /* dummy node */
	AWNode *tail; /* last node */
	AWNode *iter; /* iterator */
	int free_me;
	int cnt;
} AWList;

/**
 * @brief      loop the list
 *
 * @param      pos          AWNode *, Iterator variable
 * @param      list         List head
 */
#define AWList_for_each(pos, list)                                             \
	for ((pos) = (list)->head->next; (pos) != NULL; (pos) = (pos)->next)

/**
 * @brief      allocate memory for list
 *
 * @param      list       pointer to init, could be NULL
 *
 * @return     AWList *
 */
AWList *AWList_init(AWList *list);

/**
 * @brief      delete the whole list
 *
 * @param      list     pointer of list
 *
 */
void AWList_destroy(AWList *list);

/**
 * @brief       free node, but retain list
 */
void AWList_clear(AWList *list);

/**
 * @brief      delete node
 *
 * @param      list       pointer of list
 * @param      node       list node pointer
 *
 */
void AWList_delete_node(AWList *list, AWNode *node);

/**
 * @brief      insert node to the HEAD of list
 *
 * @param      list         pointer of list
 * @param      data         data to insert
 * @param      delete_func  function pointer to free data
 *
 */
void AWList_add_head(AWList *list, void *data, void (*delete_func)(void *));

/**
 * @brief      append node to tail of the list
 *
 * @param      list         pointer of list
 * @param      data         data to insert
 * @param      delete_func  function pointer to free data
 *
 */
void AWList_add_tail(AWList *list, void *data, void (*delete_func)(void *));

/**
 * @brief      find assigned node via callback function
 *
 * @details
 *
 * @param      list           pointer of list
 * @param      data           key data to find
 * @param      cmp_callback   callback to use data,
 *                            0: node->data is equal to the data to find
 *
 * @return     the pointer of the node if exists, NULL if not found
 */
AWNode *AWList_find_node(AWList *list, void *data,
			 int (*cmp_callback)(void *, void *));

/**
 * @brief      list iterator
 *
 * @details    always return next node after each call
 *
 * @param      list         list pointer
 *
 * @return     next node address or NULL
 */
AWNode *AWList_iterator(AWList *list);

#endif /* AWLIST_H */
